#!/usr/bin/env python3

import sys
from z3 import *
from collections import defaultdict

def get_text_file_contents_as_list (filename):
    f=open(filename,"r")
    tmp=f.readlines()
    f.close()
    return [item.rstrip() for item in tmp]

def char_is_latin(c):
    return (c>='a' and c<='z') or (c>='A' and c<='Z')

def str_is_latin(s):
    return all(char_is_latin(c) for c in s)

tmp=get_text_file_contents_as_list(sys.argv[1])
words_lst=list(filter(lambda x: str_is_latin(x), tmp))

words_total=len(words_lst)

# k=word number; v=list of chars
table=defaultdict(list)

word_n=0
for w in words_lst:
    for char in list(w.lower()):
        table[word_n].append(ord(char)-ord('a'))
    word_n+=1

# that will be a list of all chars in all words:
all_chars=set()
# k=char, v=list of words, which contains that char:
words_for_char=defaultdict(list)

for word_n in table:
    all_chars|=set(table[word_n])
    for char in table[word_n]:
        words_for_char[char].append(word_n)

# int variable for each word:
words=[Int('word_%d' % (t)) for t in range(words_total)]

# this is optimization problem, so Optimize() instead of Solver():
opt=Optimize()

# each word variable is either 0 (absent) or 1 (present):
for t in words:
    opt.add(Or(t==0, t==1))

# we know which words contain each char
# so we enumerate all words when preparing expression for each char
# we form expression like "word_1==1 OR word_2==1 OR ..." for each char:
for char in list(all_chars):
    expressions=[words[s]==1 for s in words_for_char[char]]
    # expression is a list which unfolds as list of arguments into Z3's Or() function (see asterisk)
    # that results in big expression like "Or(word_1==1, word_2==1, ...)"
    # the expression is then added as a constraint:
    opt.add(Or(*expressions))

# we need to find a such solution, where minimal number of all "word_X" variables will have 1
# "*words" unfolds to a list of arguments: [word_1, word_2, word_3,...]
# "Sum(*words)" unfolds to the following expression: "Sum(word_1, word_2, ...)"
# the sum of all "word_X" variables should be as minimal as possible:
h=opt.minimize(Sum(*words))

def print_model(m):
    result=[]
    chars=set()
    words_t=0
    for d in range(words_total):
        if m[words[d]].as_long()==1:
            result.append (words_lst[d])
            words_t+=1
            chars|=set(table[d])
    result_s=" ".join(result)
    #if len(result_s)>51:
    #    return
    print ("Result:", result_s)
    print ("Words:", words_t)
    print ("Phrase lenght (incl. spaces):", len(result_s))
    print ("Chars used:", "".join(list(map(lambda x: chr(x+ord('a')), list(chars))))) # chars used
    print ("Total chars used:", len(chars))
    print ("")
    sys.stdout.flush()

# enumerate all solutions:
results=[]
while opt.check() == sat:
    m = opt.model()

    print_model(m)

    results.append(m)
    block = []
    for d1 in m:
        t=d1()
        block.append(t != m[d1])
    opt.add(Or(block))

print ("results total=", len(results))

